import math

from error import Error
from values import *
from symbol_table import SymbolTable

class Interpreter:
    def visit(self, node, symbol_table):
        if node[0] == 'empty':
            return self.visit_empty_node(node, symbol_table)
        elif node[0] == 'number':
            return self.visit_number_node(node, symbol_table)
        elif node[0] == 'string':
            return self.visit_string_node(node, symbol_table)
        elif node[0] == 'at':
            return self.visit_at_node(node, symbol_table)
        elif node[0] == 'name':
            return self.visit_name_node(node, symbol_table)    
        elif node[0] == 'add':
            return self.visit_add_node(node, symbol_table)
        elif node[0] == 'subtract':
            return self.visit_subtract_node(node, symbol_table)
        elif node[0] == 'multiply':
            return self.visit_multiply_node(node, symbol_table)
        elif node[0] == 'divide':
            return self.visit_divide_node(node, symbol_table)
        elif node[0] == 'mod':
            return self.visit_mod_node(node, symbol_table)
        elif node[0] == 'ee':
            return self.visit_ee_node(node, symbol_table) 
        elif node[0] == 'ne':
            return self.visit_ne_node(node, symbol_table)   
        elif node[0] == 'lt':
            return self.visit_lt_node(node, symbol_table)   
        elif node[0] == 'le':
            return self.visit_le_node(node, symbol_table)   
        elif node[0] == 'gt':
            return self.visit_gt_node(node, symbol_table)    
        elif node[0] == 'ge':
            return self.visit_ge_node(node, symbol_table)    
        elif node[0] == 'and':
            return self.visit_and_node(node, symbol_table)   
        elif node[0] == 'or':
            return self.visit_or_node(node, symbol_table)   
        elif node[0] == 'xor':
            return self.visit_xor_node(node, symbol_table)    
        elif node[0] == 'plus':
            return self.visit_plus_node(node, symbol_table)
        elif node[0] == 'minus':
            return self.visit_minus_node(node, symbol_table)
        elif node[0] == 'not':
            return self.visit_not_node(node, symbol_table)
        elif node[0] == 'invert':
            return self.visit_invert_node(node, symbol_table)
        elif node[0] == 'pound':
            return self.visit_pound_node(node, symbol_table)
        elif node[0] == 'assign':
            return self.visit_assign_node(node, symbol_table)    
        elif node[0] == 'ternary':
            return self.visit_ternary_node(node, symbol_table)    
        elif node[0] == 'func':
            return self.visit_func_node(node, symbol_table)
        elif node[0] == 'call':
            return self.visit_call_node(node, symbol_table)
        elif node[0] == 'statements':
            return self.visit_statements_node(node, symbol_table)
        else:
            raise Exception(f'Unknown node type')

    def visit_empty_node(self, node, symbol_table):
        return At()

    def visit_number_node(self, node, symbol_table):
        return Number(node[1])

    def visit_string_node(self, node, symbol_table):
        return String(node[1])

    def visit_at_node(self, node, symbol_table):
        return At()

    def visit_name_node(self, node, symbol_table):
        name = node[1]

        result = symbol_table.get(name)

        if result is None:
            raise Error(f'{name} is not defined')
        
        return result

    def visit_add_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).add(self.visit(node[2], symbol_table))
    
    def visit_subtract_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).subtract(self.visit(node[2], symbol_table))

    def visit_multiply_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).multiply(self.visit(node[2], symbol_table))

    def visit_divide_node(self, node, symbol_table):
        try:
            return self.visit(node[1], symbol_table).divide(self.visit(node[2], symbol_table))
        except ZeroDivisionError:
            return Number(math.inf)

    def visit_mod_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).mod(self.visit(node[2], symbol_table))

    def visit_ee_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).eq(self.visit(node[2], symbol_table))

    def visit_ne_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).ne(self.visit(node[2], symbol_table))

    def visit_lt_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).lt(self.visit(node[2], symbol_table))

    def visit_gt_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).gt(self.visit(node[2], symbol_table))

    def visit_le_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).le(self.visit(node[2], symbol_table))

    def visit_ge_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).ge(self.visit(node[2], symbol_table))

    def visit_and_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).and_(self.visit(node[2], symbol_table))

    def visit_or_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).or_(self.visit(node[2], symbol_table))

    def visit_xor_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).xor_(self.visit(node[2], symbol_table))
        
    def visit_plus_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).plus()

    def visit_minus_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).minus()

    def visit_not_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).not_()

    def visit_invert_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).invert()

    def visit_pound_node(self, node, symbol_table):
        return self.visit(node[1], symbol_table).pound()

    def visit_assign_node(self, node, symbol_table):
        if node[1][0] != 'name':
            raise Error('Invalid left-hand side in assignment')

        name = node[1][1]
        value = self.visit(node[2], symbol_table)

        symbol_table.set(name, value)
    
        return value

    def visit_ternary_node(self, node, symbol_table):
        condition = self.visit(node[1], symbol_table)

        if not isinstance(condition, Number):
            raise Error(f'{condition} is not a number')
    
        if condition.value:
            return self.visit(node[2], symbol_table)
        else:
            return self.visit(node[3], symbol_table)

    def visit_func_node(self, node, symbol_table):
        def func(args):
            new_symbol_table = SymbolTable(symbol_table)

            for i in range(0, len(args)):
                new_symbol_table.set(f'${i}', args[i])
            
            return self.visit(node[1], new_symbol_table)
            
        return Func(func)

    def visit_call_node(self, node, symbol_table):
        func = self.visit(node[1], symbol_table)
        args = []

        for arg_node in node[2]:
            args.append(self.visit(arg_node, symbol_table))

        return func.func(args)

    def visit_statements_node(self, node, symbol_table):
        for i in range(1, len(node) - 1):
            self.visit(node[i], symbol_table)

        return self.visit(node[-1], symbol_table)
 


#!/usr/bin/python3

import sys

from q import run, run_text

if len(sys.argv) == 1:
    while True:       
        text = input('> ')
        if text == 'exit': break
        print(run_text(text))
elif len(sys.argv) == 2:
    run(sys.argv[1])
else:
    print('Usage: q [file]')


import string

from error import Error

WHITESPACE = ' \t'
LETTERS = string.ascii_letters + '_$'
DIGITS = '0123456789'

class Lexer:
    def __init__(self, text):
        self.text = iter(text)
        self.advance()
    
    def advance(self):
        try:
            self.current_char = next(self.text)
        except StopIteration:
            self.current_char = None
    
    def get_tokens(self):
        tokens = []

        while self.current_char != None:
            if self.current_char in WHITESPACE:
                self.advance()
            elif self.current_char == ';' or self.current_char == '\n':
                self.advance()
                tokens.append(('newline',))
            elif self.current_char == '.' or self.current_char in DIGITS:
                tokens.append(self.get_number())
            elif self.current_char in LETTERS:
                tokens.append(self.get_name())
            elif self.current_char == '"':
                tokens.append(self.get_string())
            elif self.current_char == '+':
                self.advance()
                tokens.append(('plus',))
            elif self.current_char == '-':
                self.advance()
                tokens.append(('minus',))
            elif self.current_char == '*':
                self.advance()
                tokens.append(('multiply',))
            elif self.current_char == '/':
                self.advance()
                tokens.append(('divide',))
            elif self.current_char == '%':
                self.advance()
                tokens.append(('mod',))
            elif self.current_char == '<':
                tokens.append(self.get_lt())
            elif self.current_char == '=':
                tokens.append(self.get_eq())
            elif self.current_char == '>':
                tokens.append(self.get_gt())
            elif self.current_char == '&':
                self.advance()
                tokens.append(('and',))
            elif self.current_char == '|':
                self.advance()
                tokens.append(('or',)) 
            elif self.current_char == '^':
                self.advance()
                tokens.append(('xor',))
            elif self.current_char == '!':
                tokens.append(self.get_not())  
            elif self.current_char == '~':
                self.advance()
                tokens.append(('invert',))
            elif self.current_char == '?':
                self.advance()
                tokens.append(('question',))     
            elif self.current_char == ':':
                self.advance()
                tokens.append(('colon',))  
            elif self.current_char == ',':
                self.advance()
                tokens.append(('comma',))  
            elif self.current_char == '#':
                self.advance()
                tokens.append(('pound',)) 
            elif self.current_char == '@':
                self.advance()
                tokens.append(('at',))                              
            elif self.current_char == '(':
                self.advance()
                tokens.append(('lparen',))
            elif self.current_char == ')':
                self.advance()
                tokens.append(('rparen',))
            elif self.current_char == '{':
                self.advance()
                tokens.append(('lcurly',)) 
            elif self.current_char == '}':
                self.advance()
                tokens.append(('rcurly',)) 
            else:
                raise Error('Lexical error')

        tokens.append(('eof',))

        return tokens

    def get_number(self):
        decimal_point_count = 0
        number_str = self.current_char
        self.advance()

        while self.current_char != None and (self.current_char == '.' or self.current_char in DIGITS):
            if self.current_char == '.':
                decimal_point_count += 1
                if decimal_point_count > 1:
                    break
            
            number_str += self.current_char
            self.advance()

        if number_str.startswith('.'):
            number_str = '0' + number_str
        if number_str.endswith('.'):
            number_str += '0'

        return ('number', float(number_str))

    def get_name(self):
        name_str = self.current_char
        self.advance()

        while self.current_char != None and (self.current_char in LETTERS or self.current_char in DIGITS):
            name_str += self.current_char
            self.advance()

        return ('name', name_str)
    
    def get_string(self):   
        string = ''
        escape_character = False
        self.advance()

        escape_characters = {
            'n': '\n',
            't': '\t'
        }

        while self.current_char != None and (self.current_char != '"' or escape_character):
            if escape_character:
                string += escape_characters.get(self.current_char, self.current_char)
            else:
                if self.current_char == '\\':
                    escape_character = True
                else:
                    string += self.current_char
                    escape_character = False
                    
            self.advance()
        
        self.advance()
        
        return ('string', string)

    def get_lt(self):
        self.advance()
        if self.current_char == '=':
            self.advance()
            return ('le',)
        else:
            return ('lt',)

    def get_eq(self):
        self.advance()
        if self.current_char == '=':
            self.advance()
            return ('ee',)
        else:
            return ('eq',)

    def get_gt(self):
        self.advance()
        if self.current_char == '=':
            self.advance()
            return ('ge',)
        else:
            return ('gt',)

    def get_not(self):
        self.advance()
        if self.current_char == '=':
            self.advance()
            return ('ne',)
        else:
            return ('not',)



from error import Error

class Parser:
    def __init__(self, tokens):
        self.tokens = iter(tokens)
        self.advance()

    def raise_error(self):
        raise Error('Syntax error')

    def advance(self):
        try:
            self.current_token = next(self.tokens)
        except StopIteration:
            self.current_token = None

    def parse(self):
        if self.current_token[0] == 'eof':
            return ('empty',)

        result = self.statements()

        if self.current_token[0] != 'eof':
            self.raise_error()

        return result

    def statements(self):
        statements = []

        statement = self.expr()
        statements.append(statement)

        while self.current_token[0] != 'eof' and self.current_token[0] == 'newline':
            self.advance()
            statements.append(self.expr())

        return ('statements', *statements)

    def expr(self):
        if self.current_token[0] == 'lcurly':
            return self.func_expr()

        return self.assignment_expr()

    def func_expr(self):
        self.advance()

        if self.current_token[0] == 'rcurly':
            self.advance()
            return ('func', ('empty',))

        body = self.statements()

        if self.current_token[0] != 'rcurly':
            self.raise_error()

        self.advance()
        return ('func', body)

    def assignment_expr(self):
        result = self.ternary_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] == 'eq':
            self.advance()
            result = ('assign', result, self.expr())

        return result

    def ternary_expr(self):
        result = self.xor_expr()
        
        if self.current_token[0] == 'question':
            self.advance()

            if_expr = self.expr()
            
            if self.current_token[0] != 'colon':
                self.raise_error()

            self.advance()
            else_expr = self.expr()

            return ('ternary', result, if_expr, else_expr)
        
        return result

    def xor_expr(self):
        result = self.or_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] == 'xor':
            self.advance()
            result = ('xor', result, self.or_expr())

        return result

    def or_expr(self):
        result = self.and_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] == 'or':
            self.advance()
            result = ('or', result, self.and_expr())

        return result

    def and_expr(self):
        result = self.relational_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] == 'and':
            self.advance()
            result = ('and', result, self.relational_expr())

        return result

    def relational_expr(self):
        result = self.equality_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] in ('lt', 'gt', 'le', 'ge'):
            if self.current_token[0] == 'lt':
                self.advance()
                result = ('lt', result, self.equality_expr())
            elif self.current_token[0] == 'gt':
                self.advance()
                result = ('gt', result, self.equality_expr())
            elif self.current_token[0] == 'le':
                self.advance()
                result = ('le', result, self.equality_expr())
            elif self.current_token[0] == 'ge':
                self.advance()
                result = ('ge', result, self.equality_expr())

        return result

    def equality_expr(self):
        result = self.additive_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] in ('ee', 'ne'):
            if self.current_token[0] == 'ee':
                self.advance()
                result = ('ee', result, self.additive_expr())
            elif self.current_token[0] == 'ne':
                self.advance()
                result = ('ne', result, self.additive_expr())

        return result

    def additive_expr(self):
        result = self.term()

        while self.current_token[0] != 'eof' and self.current_token[0] in ('plus', 'minus'):
            if self.current_token[0] == 'plus':
                self.advance()
                result = ('add', result, self.term())
            elif self.current_token[0] == 'minus':
                self.advance()
                result = ('subtract', result, self.term())

        return result

    def term(self):
        result = self.call_expr()

        while self.current_token[0] != 'eof' and self.current_token[0] in ('multiply', 'divide', 'mod'):
            if self.current_token[0] == 'multiply':
                self.advance()
                result = ('multiply', result, self.call_expr())
            elif self.current_token[0] == 'divide':
                self.advance()
                result = ('divide', result, self.call_expr())
            elif self.current_token[0] == 'mod':
                self.advance()
                result = ('mod', result, self.call_expr())
                
        return result

    def call_expr(self):
        factor = self.factor()

        while self.current_token[0] == 'lparen':
            self.advance()

            arg_nodes = []

            if self.current_token[0] == 'rparen':
                self.advance()
                return ('call', factor, ())
            else:
                arg_nodes.append(self.expr())

                while self.current_token[0] == 'comma':
                    self.advance()
                    arg_nodes.append(self.expr())

                self.advance()

                factor = ('call', factor, arg_nodes)
            
        return factor
                    
    def factor(self):
        token = self.current_token

        if token[0] == 'lparen':
            self.advance()
            result = self.expr()

            if self.current_token[0] != 'rparen':
                self.raise_error()
            
            self.advance()
            return result

        elif token[0] == 'number':
            self.advance()
            return ('number', token[1])

        elif token[0] == 'string':
            self.advance()
            return ('string', token[1])

        elif token[0] == 'at':
            self.advance()
            return ('at',)

        elif token[0] == 'name':
            self.advance()
            return ('name', token[1])

        elif token[0] == 'plus':
            self.advance()
            return ('plus', self.factor())
        
        elif token[0] == 'minus':
            self.advance()
            return ('minus', self.factor())

        elif token[0] == 'not':
            self.advance()
            return ('not', self.factor())

        elif token[0] == 'invert':
            self.advance()
            return ('invert', self.factor())

        elif token[0] == 'pound':
            self.advance()
            return ('pound', self.factor())

        self.raise_error()


from symbol_table import SymbolTable
from values import *

def func_I(args):
    return String(input())

# def func_N(args):
#     return Number(float(args[0]))

def func_S(args):
    return String(str(args[0]))

def func_P(args):
    print(args[0])
    return At()

stdlib = SymbolTable()
stdlib.set('I', Func(func_I))
stdlib.set('S', Func(func_S))
# stdlib.set('N', Func(func_N))
stdlib.set('P', Func(func_P))